.MODULE/BOOT=0  fft;

{       Performs 32-bit Radix-2 DIT FFT (32-bit data, 16-bit twiddle factors)

        Calling Parameters
                real_lsw = Real input data in scrambled order
                real_msw = Real input data 16-bit sign extention
                imag_lsw = Imaginary input data in scrambled order
                imag_msw = Imaginary input data 16-bit sign extention
                twid_real = Twiddle factor cosine values
                twid_imag = Twiddle factor sine values
                groups = N/2
                bflys_per_group = 1
                node_space = 1

        Return Values
                real_lsw = LSW of real results in normal order
                real_msw = MSW of real results in normal order
                imag_lsw = LSW of imaginary results in normal order
                imag_msw = MSW of imaginary results in normal order

        Altered Registers
                I0,I1,I2,I3,I4,I5,L0,L1,L2,L3,L4,L5
                M0,M1,M2,M3,M4,M5
                AX0,AX1,AY0,AY1,AR,AF
                MX0,MX1,MY0,MY1,MR,SB,SE,SR,SI

        Altered Memory
                real_lsw, real_msw, imag_lsw, imag_msw, groups,
                node_space, bflys_per_group, blk_exponent

        Computation Time
                10 + [((22*N/2)+18)*log2(N)] + (N-1)*10

                = 11262 cycles for N = 128
}

.CONST    log2N=7, N=128;       {Set constants for N-point FFT}
.EXTERNAL twid_real, twid_imag;
.EXTERNAL real_lsw, real_msw, imag_lsw, imag_msw;
.EXTERNAL groups, bflys_per_group, node_space, blk_exponent;
.ENTRY    fft_strt;

fft_strt: CNTR=log2N;     {Initialize stage counter}
          M0=-N+1;
          M1=1;
          M3=N;
          L0=N*2; L1=N*2; L2=N*2; L3=N*2;
          L4=%twid_real;
          L5=%twid_imag;
          DO stage_loop UNTIL CE;   {Compute all stages in FFT}
            I0=^real_lsw;           {I0 -->x0_lsw in 1st grp of stage}
            I2=^imag_lsw;           {I2 -->y0_lsw in 1st grp of stage}
            SI=DM(groups);
            CNTR=SI;                {CNTR = group counter}
            M4=SI;                  {M4=twiddle factor modifier}
            M2=DM(node_space);      {M2=node space modifier}
            I1=I0;
            MODIFY(I1,M2);          {I1 -->x1_lsw of 1st grp in stage}
            I3=I2;
            MODIFY(I3,M2);          {I3 -->y1_lsw of 1st grp in stage}
            DO group_loop UNTIL CE;
              I4=^twid_real;               {I4 --> C of W0}
              I5=^twid_imag;               {I5 --> (-S) of W0}
              MY0=PM(I4,M4),MX1=DM(I3,M3); {MY0=C,MX1=y1_lsw }
              SR1=DM(I3,M3);               {SR1=y1_msw}
              CNTR=DM(bflys_per_group);    {CNTR = butterfly counter}
              DO bfly_loop UNTIL CE;
                MR=MX1*MY0(US),MY1=PM(I5,M4),MX0=DM(I1,M3);
                                                {MR=y1_lsw(C)}
                                                {MY0=-S,MX1=x1_lsw }
                MR=MR+MX0*MY1(US),SR0=DM(I1,M3);{accumulate x1_lsw(-S)}
                                                {SR1=x1_msw}
                MR0=MR1;                        {shift accumulator 16-bits}
                MR1=MR2;
                MR=MR+SR0*MY1(SS),AY0=DM(I2,M3);{accumulate x1_msw(-S)}
                                                {AY0=y0_lsw}
                MR=MR+SR1*MY0(SS),AY1=DM(I2,M3);{accumulate y1_msw(C)}
                                                {AY1=y0_msw}
                AR=AY0-MR0;                     {y1_lsw=y0_lsw-accumulator}
                DM(I3,M3)=AR, AR=AY1-MR1+C-1;   {store new y1_lsw}
                                                {y1_msw=y0_msw-accumulator}
                DM(I3,M0)=AR, AR=MR0+AY0;       {store new y1_msw}
                                                {y0_lsw=y0_lsw+accumulator}
                DM(I2,M3)=AR, AR=MR1+AY1+C;     {store new y0_lsw}
                                                {y0_msw=y0_msw+accumulator}
                DM(I2,M0)=AR;                   {store new y0_msw}
                MR=MX0*MY0(US),AY0=DM(I0,M3);   {MR=x1_lsw(C),AY0=x0_lsw}
                MR=MR-MX1*MY1(US),MX1=DM(I3,M3);{acc/subtract y1_lsw(-S)}
                                                {MX1=next y1_lsw}
                MR0=MR1;                        {shift accumulator 16-bits}
                MR1=MR2;
                MR=MR+SR0*MY0(SS),AY1=DM(I0,M3);{accumulate x1_msw(C)}
                                                {AY1=x0_msw}
                MR=MR-SR1*MY1(SS),SR1=DM(I3,M3);{accumulate y1_msw(-S)}
                                                {SR1=y1_msw}
                AR=AY0-MR0,MY0=PM(I4,M4);       {x1_lsw=x0_lsw-accumulator}
                                                {MY0=next C}
                DM(I1,M3)=AR, AR=AY1-MR1+C-1;   {store new x1_lsw}
                                                {x1_msw=x0_msw-accumulator}
                DM(I1,M0)=AR, AR=MR0+AY0;       {store new x1_msw}
                                                {x0_lsw=x0_lsw+accumulator}
                DM(I0,M3)=AR, AR=MR1+AY1+C;     {store new x0_lsw}
                                                {x0_msw=x0_msw+accumulator}
bfly_loop:      DM(I0,M0)=AR;                   {store new x0_msw}


              MODIFY(I0,M2);  {I0 -->1st x0 in next group}
              MODIFY(I1,M2);  {I1 -->1st x1 in next group}
              MODIFY(I2,M2);  {I2 -->1st y0 in next group}
group_loop:   MODIFY(I3,M2);  {I3 -->1st y1 in next group}
            SI=DM(bflys_per_group);
            SR=ASHIFT SI BY 1(LO);
            DM(node_space)=SR0;         {node_space=node_space*2}
            DM(bflys_per_group)=SR0;    {bflys_per_group=bflys_per_group*2}
            SI=DM(groups);
            SR=ASHIFT SI BY -1(LO);
stage_loop: DM(groups)=SR0;             {groups=groups/2}
        RTS;
.ENDMOD;